/* $Id: gpio.c,v 1.2 2001/12/08 07:48:07 nari Exp $ */

#define MODULE
#define __KERNEL__

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/mm.h> 
#include <asm/uaccess.h>
#include <asm/vr41xx.h>
#include <asm/io.h>  

#ifndef GPIO_MAJOR
#define GPIO_MAJOR	17
#endif

int gpio_major = GPIO_MAJOR;

static char gpio_bit_usable[32]=
{
  1,1,1,0, 0,1,1,0, 1,0,0,1, 1,1,1,1,   /* GPIO  0-15 */
  0,0,0,1, 1,0,1,1, 1,0,0,0, 0,1,0,1,   /* GPIO 16-31 */    
};

typedef struct _gpio_cont {
  int bit;
  int bBinary;
} GpioCont;

void gpio_free_interrupt(GpioCont *cont);

static void regbset(volatile unsigned short *port,unsigned mask,unsigned bits)
{
  unsigned short data0,data;
  data0 = readw(port);
  data = (data0 & ~mask) | (bits & mask);
  printk(KERN_DEBUG "regbset [%p] %04x->%04x\n",port,data0,data);
  writew(data,port);
}

int gpio_open(struct inode *inode, struct file *filp)
{
  int minor = MINOR(inode->i_rdev);
  int bit = minor & 0x1f;
  GpioCont *cont;
  
  if(!gpio_bit_usable[bit])
    return -NODEV; 
  cont = (GpioCont *)kmalloc(sizeof(GpioCont), GFP_KERNEL);
  cont->bit = bit;
  cont->bBinary = minor & 0x20;
  filp->private_data = cont;
  MOD_INC_USE_COUNT;
  return 0;
}

int gpio_release(struct inode *inode, struct file *filp)
{
  GpioCont *cont = (GpioCont *)filp->private_data;
  kfree(cont);
  MOD_DEC_USE_COUNT;
  return 0;
}

ssize_t gpio_read(struct file *filp, char *buf, size_t count,loff_t *f_pos)
{
  GpioCont *cont = (GpioCont *)filp->private_data;
  int bit = cont->bit;
  volatile unsigned short *mode_port = VR41XX_GPMD0REG + (bit / 8);
  volatile unsigned short *data_port = VR41XX_GPDATLREG - (bit / 16);
  unsigned short mode,data,mask;
  ssize_t r;
  
  mask = 3 << ((bit % 8) * 2);
  regbset(mode_port,mask, 0);  /* input */
  mask = 1 << (bit % 16);
  
  if(cont->bBinary){
    int i;
    if(!access_ok(VERIFY_WRITE, (void *)buf, count)) 
      return -EFAULT;
    for(i=0;i<count;i++){
      data = readw(data_port);
      __put_user((data & mask) ? 1 : 0, buf+i);
    }
    r = count;
  }
  else {
    if(count < 2) {
      printk(KERN_ERR "gpio_read buffer size must be greater than 1\n");
      return -EINVAL;
    }
    data = readw(data_port);
    if(put_user((data & mask) ? '1' : '0', buf)) return -EFAULT;
    if(put_user('\n', buf+1)) return -EFAULT;
    r = 2;
  }
  *f_pos += r;
  return r;
}

ssize_t gpio_write(struct file *filp, const char *buf,size_t count,
		   loff_t *f_pos)
{
  GpioCont *cont = (GpioCont *)filp->private_data;
  int bit     = cont->bit;
  volatile unsigned short *mode_port = VR41XX_GPMD0REG + (bit / 8);
  volatile unsigned short *data_port = VR41XX_GPDATLREG - (bit / 16);
  unsigned short mode,data,data1,mask;
  ssize_t r;
  int i;
  
  mask = 3 << ((bit % 8) * 2);
  data = 2 << ((bit % 8) * 2);
  regbset(mode_port, mask, data);  /* output mode */
  
  mask = 1 << (bit % 16);
  data = readw(data_port);
  data &= ~mask;

  for(i=0;i<count;i++){
    int c;
    data1 = data;
    if(get_user(c, buf+i)) return -EFAULT;
    if(c==' ' || c=='\t' || c=='\n' || c=='\r') continue;
    else if(c==1 || c=='1' || c == 'y' || c == 'Y' || c == 't' || c == 'T')
      data1 |= mask;
    writew(data1,data_port);
  }
  *f_pos += count;
  return count;
}

struct file_operations gpio_fops = {
  read:    gpio_read,  
  write:   gpio_write, 
  open:    gpio_open,
  release: gpio_release,
};

int init_module(void)
{
  int r;
  r = register_chrdev(gpio_major, "gpio", &gpio_fops);
  if(r < 0){
    printk(KERN_WARNING "gpio: can't get major %d\n", gpio_major);
    return r;
  }
  if(gpio_major == 0) gpio_major = r;
  printk(KERN_INFO "GPIO installed\n");
  return 0;
}

void cleanup_module(void)
{
  int r;
  r = unregister_chrdev(gpio_major, "gpio");
  if(r < 0){
    printk(KERN_ALERT "gpio: unregister major %d failed\n", gpio_major);
    return;
  }
  printk(KERN_INFO "gpio uninstalled\n");
}
